# Copyright European Organization for Nuclear Research (CERN)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Luis Rodrigues, <luis.rodrigues@cern.ch>, 2013


# client2.add_replica(rse_src, scope_name, did_name, did_bytes)
# client2.add_replication_rule(dids=[{'scope': scope_name, 'name': did_name}], account=account_name, copies=2, rse_expression='%s|%s'%(rse1, rse2), grouping='NONE')
