
# This file is automatically generated by setup.py, So don't edit it. :)
VERSION_INFO = {
    'final': True,
    'version': '1.5.11.post1',
    'branch_nick': '(no',
    'revision_id': 'f84b07634b0a0e454febcc0466e7882316593adf',
    'revno': 3945
}
